import os
import pathlib

import pytest

import xorq.api as xo
from xorq.common.utils.env_utils import maybe_substitute_env_vars
from xorq.vendor.ibis.backends import BaseBackend
from xorq.vendor.ibis.backends.profiles import Profile, Profiles


local_con_names = ("duckdb", "xorq", "datafusion", "pandas", "pyiceberg")
remote_connectors = (lambda: xo.postgres.connect_env(),)
local_connectors = tuple(
    lambda: getattr(xo, con_name).connect()  # noqa: B023
    for con_name in local_con_names
)


@pytest.mark.parametrize(
    "con_name",
    local_con_names
    + (
        pytest.param(
            "invalid-con-name", marks=pytest.mark.xfail(reason="only valid con names")
        ),
    ),
)
def test_con_has_profile(con_name):
    con = getattr(xo, con_name).connect()
    assert isinstance(con, BaseBackend)
    profile = getattr(con, "_profile", None)
    assert isinstance(profile, Profile)
    assert profile.almost_equals(Profile.from_con(con))
    #
    other = profile.get_con()
    assert con.name == other.name
    # this doesn't work because _con_args, _con_kwargs doesn't get the defaults which are eventually invoked
    # assert hash(con) == hash(other)
    assert profile.almost_equals(other._profile)


@pytest.mark.postgres
@pytest.mark.parametrize("connect", remote_connectors)
def test_remote_con_works(connect):
    con = connect()
    assert isinstance(con, BaseBackend)
    profile = getattr(con, "_profile", None)
    assert isinstance(profile, Profile)
    assert profile.almost_equals(Profile.from_con(con))
    #
    other = profile.get_con()
    assert con.name == other.name
    # this doesn't work because _con_args, _con_kwargs doesn't get the defaults which are eventually invoked
    # assert hash(con) == hash(other)
    assert profile.almost_equals(other._profile)
    assert con.list_tables() == other.list_tables()


def test_profiles(monkeypatch, tmp_path):
    default_profile_dir = xo.options.profiles.profile_dir
    assert default_profile_dir == pathlib.Path("~/.config/xorq/profiles").expanduser()
    profiles = Profiles()
    assert profiles.profile_dir == default_profile_dir

    monkeypatch.setattr(xo.options.profiles, "profile_dir", tmp_path)
    profiles = Profiles()
    assert profiles.profile_dir == tmp_path
    assert not profiles.list()


@pytest.mark.parametrize(
    "connector",
    [pytest.param(c, marks=pytest.mark.postgres) for c in remote_connectors]
    + list(local_connectors),
)
def test_save_load(connector, monkeypatch, tmp_path):
    monkeypatch.setattr(xo.options.profiles, "profile_dir", tmp_path)
    # In the is case letsql has a raw passwords so its value is
    # ***elided*** so we can't instantiate it
    os.environ["LETSQL_PASSWORD"] = "letsql"
    con = connector()
    profiles = Profiles()
    profile = con._profile
    profile.save(check_secrets=False)

    others = (
        profiles.get(profile.hash_name),
        profiles[profile.hash_name],
        profile.load(profile.hash_name),
    )

    for other in others:
        assert profile == other
        assert con.list_tables() == other.get_con().list_tables()

    del os.environ["LETSQL_PASSWORD"]


def test_profile_hash_order_independence():
    # different order same kwargs
    profile1 = Profile(
        con_name="postgres",
        kwargs_tuple=(
            ("host", "localhost-dum"),
            ("port", 5432),
            ("user", "testuser"),
            ("database", "testdb"),
        ),
    )

    profile2 = Profile(
        con_name="postgres",
        kwargs_tuple=(
            ("port", 5432),
            ("database", "testdb"),
            ("user", "testuser"),
            ("host", "localhost-dum"),
        ),
    )

    # check hash order agnostic
    assert profile1.hash_name.split("_")[0] == profile2.hash_name.split("_")[0]
    assert profile1.hash_name.split("_")[1] != profile2.hash_name.split("_")[1]
    # check sort
    assert profile1.kwargs_tuple == profile2.kwargs_tuple

    # check clone
    cloned = profile1.clone()
    assert cloned.hash_name.split("_")[0] == profile1.hash_name.split("_")[0]


class TestParseEnvVars:
    def test_empty_dict(self):
        """Test with empty dictionary."""
        assert maybe_substitute_env_vars({}) == {}

    def test_no_env_vars(self):
        """Test with dictionary containing no environment variables."""
        input_dict = {
            "host": "localhost",
            "port": 5432,
            "user": "postgres",
            "non_string": 123,
            "none_value": None,
            "empty_string": "",
        }
        assert maybe_substitute_env_vars(input_dict) == input_dict

    def test_dollar_brace_format(self, monkeypatch):
        """Test with ${VAR} format environment variables."""
        # Set environment variables for testing
        monkeypatch.setenv("TEST_USER", "testuser")
        monkeypatch.setenv("TEST_PASSWORD", "secretpass")

        input_dict = {
            "host": "localhost",
            "port": 5432,
            "user": "${TEST_USER}",
            "password": "${TEST_PASSWORD}",
            "non_env": "regular_value",
        }

        expected = {
            "host": "localhost",
            "port": 5432,
            "user": "testuser",
            "password": "secretpass",
            "non_env": "regular_value",
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_dollar_format(self, monkeypatch):
        """Test with $VAR format environment variables."""
        # Set environment variables for testing
        monkeypatch.setenv("TEST_USER", "testuser")
        monkeypatch.setenv("TEST_PASSWORD", "secretpass")

        input_dict = {
            "host": "localhost",
            "port": 5432,
            "user": "$TEST_USER",
            "password": "$TEST_PASSWORD",
            "non_env": "regular_value",
        }

        expected = {
            "host": "localhost",
            "port": 5432,
            "user": "testuser",
            "password": "secretpass",
            "non_env": "regular_value",
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_mixed_formats(self, monkeypatch):
        """Test with mixed ${VAR} and $VAR formats."""
        # Set environment variables for testing
        monkeypatch.setenv("TEST_USER", "testuser")
        monkeypatch.setenv("TEST_PASSWORD", "secretpass")

        input_dict = {
            "host": "localhost",
            "port": 5432,
            "user": "${TEST_USER}",
            "password": "$TEST_PASSWORD",
            "non_env": "regular_value",
        }

        expected = {
            "host": "localhost",
            "port": 5432,
            "user": "testuser",
            "password": "secretpass",
            "non_env": "regular_value",
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_non_string_values(self, monkeypatch):
        """Test with non-string values."""
        monkeypatch.setenv("TEST_VAR", "test_value")

        input_dict = {
            "string": "${TEST_VAR}",
            "integer": 123,
            "float": 45.67,
            "boolean": True,
            "none": None,
            "list": [1, 2, 3],
            "dict": {"a": 1, "b": 2},
        }

        expected = {
            "string": "test_value",
            "integer": 123,
            "float": 45.67,
            "boolean": True,
            "none": None,
            "list": [1, 2, 3],
            "dict": {"a": 1, "b": 2},
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_missing_env_var(self, monkeypatch):
        """Test with missing environment variables - should raise ValueError."""
        monkeypatch.setenv("EXISTING_VAR", "exists")

        input_dict = {
            "existing": "${EXISTING_VAR}",
            "missing": "${MISSING_VAR}",
            "regular": "value",
        }

        with pytest.raises(KeyError, match="'MISSING_VAR'"):
            maybe_substitute_env_vars(input_dict)

    def test_dollar_sign_in_string(self, monkeypatch):
        """Test with strings containing dollar signs but not as env vars."""
        input_dict = {
            "code": "a$b$c",
            "text": "This costs $5",
            "complex": "a${non-env}b",  # Not a proper env var syntax
        }

        expected = {
            "code": "a$b$c",
            "text": "This costs $5",
            "complex": "a${non-env}b",
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_preserve_case(self, monkeypatch):
        """Test that environment variable case is preserved."""
        monkeypatch.setenv("UPPERCASE", "value1")
        monkeypatch.setenv("lowercase", "value2")
        monkeypatch.setenv("MixedCase", "value3")

        input_dict = {
            "var1": "${UPPERCASE}",
            "var2": "${lowercase}",
            "var3": "${MixedCase}",
        }

        expected = {
            "var1": "value1",
            "var2": "value2",
            "var3": "value3",
        }

        assert maybe_substitute_env_vars(input_dict) == expected

    def test_nested_structures(self, monkeypatch):
        """Test how function handles nested structures (should only process top level)."""
        monkeypatch.setenv("TEST_VAR", "test_value")

        input_dict = {
            "top_level": "${TEST_VAR}",
            "nested_dict": {
                "env_var": "${TEST_VAR}",  # This should not be processed
                "normal": "value",
            },
            "list_with_vars": ["${TEST_VAR}", "normal"],  # This should not be processed
        }

        expected = {
            "top_level": "test_value",
            "nested_dict": {"env_var": "${TEST_VAR}", "normal": "value"},
            "list_with_vars": ["${TEST_VAR}", "normal"],
        }

        assert maybe_substitute_env_vars(input_dict) == expected


def test_connection_with_env_vars_preserves_env_vars(monkeypatch, tmp_path):
    """Test that connections instantiated with env vars preserve them in profiles."""

    # Set up test environment
    monkeypatch.setattr(xo.options.profiles, "profile_dir", tmp_path)

    # Set environment variables to match the existing profile values
    monkeypatch.setenv("POSTGRES_HOST", "localhost")
    monkeypatch.setenv("POSTGRES_USER", "postgres")
    monkeypatch.setenv("POSTGRES_PASSWORD", "postgres")
    monkeypatch.setenv("POSTGRES_DB", "ibis_testing")

    con_postgres = xo.postgres.connect(
        host="${POSTGRES_HOST}",
        user="${POSTGRES_USER}",
        password="${POSTGRES_PASSWORD}",
        database="${POSTGRES_DB}",
        port=5432,
    )

    # Get profile from connection
    profile = con_postgres._profile

    # Verify profile has env var references
    assert profile.kwargs_dict["host"] == "${POSTGRES_HOST}"
    assert profile.kwargs_dict["user"] == "${POSTGRES_USER}"
    assert profile.kwargs_dict["password"] == "${POSTGRES_PASSWORD}"
    assert profile.kwargs_dict["database"] == "${POSTGRES_DB}"

    # Save profile
    profile.save(alias="pg_env_var_test", check_secrets=False)

    # Create Profiles instance to load profiles
    profiles = Profiles(profile_dir=tmp_path)

    # Get profiles in different ways
    loaded_profiles = [
        profiles.get(profile.hash_name),
        profiles[profile.hash_name],
        profile.load(profile.hash_name, profile_dir=tmp_path),
    ]

    # Verify all loaded profiles have env var references
    for loaded_profile in loaded_profiles:
        assert loaded_profile.kwargs_dict["host"] == "${POSTGRES_HOST}"
        assert loaded_profile.kwargs_dict["user"] == "${POSTGRES_USER}"
        assert loaded_profile.kwargs_dict["password"] == "${POSTGRES_PASSWORD}"

        # Create connection from loaded profile
        loaded_con = loaded_profile.get_con()

        # Verify the connection's profile still has env var references
        assert loaded_con._profile is not None
        assert loaded_con._profile.kwargs_dict["host"] == "${POSTGRES_HOST}"
        assert loaded_con._profile.kwargs_dict["user"] == "${POSTGRES_USER}"
        assert loaded_con._profile.kwargs_dict["password"] == "${POSTGRES_PASSWORD}"

        # Test that the connection works by comparing to a simple list_tables call
        tables1 = con_postgres.list_tables()
        tables2 = loaded_con.list_tables()
        assert tables1 == tables2


class TestCheckForExposedSecrets:
    def test_password_no_env_var(self):
        """Test that a profile with password not using env var is rejected."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "secret"),  # Not using env var
            ),
        )

        with pytest.raises(ValueError) as excinfo:
            profile.check_for_exposed_secrets()

        # Check error message contains password
        assert "'password'" in str(excinfo.value)
        assert "$password or ${password}" in str(excinfo.value)

    def test_password_with_env_var_dollar(self):
        """Test that a profile with password using $ENV_VAR format is accepted."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "$PASSWORD"),  # Using env var
            ),
        )

        # Should not raise an error
        profile.check_for_exposed_secrets()

    def test_password_with_env_var_dollar_brace(self):
        """Test that a profile with password using ${ENV_VAR} format is accepted."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "${PASSWORD}"),  # Using env var
            ),
        )

        # Should not raise an error
        profile.check_for_exposed_secrets()

    def test_postgres_specific_secret_keys(self):
        """Test that postgres-specific secret keys are checked."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "$PASSWORD"),  # Using env var
                ("sslcert", "/path/to/cert"),  # Not using env var
            ),
        )

        with pytest.raises(ValueError) as excinfo:
            profile.check_for_exposed_secrets()

        # Check error message contains sslcert
        assert "'sslcert'" in str(excinfo.value)

    def test_snowflake_specific_secret_keys(self):
        """Test that snowflake-specific secret keys are checked."""
        profile = Profile(
            con_name="snowflake",
            kwargs_tuple=(
                ("host", "localhost"),
                ("database", "snowflake"),
                ("password", "$PASSWORD"),  # Using env var
                (
                    "user",
                    "snowuser",
                ),  # Not using env var - snowflake treats this as sensitive
            ),
        )

        with pytest.raises(ValueError) as excinfo:
            profile.check_for_exposed_secrets()

        # Check error message contains user
        assert "'user'" in str(excinfo.value)

    def test_check_secrets_disabled(self, tmp_path):
        """Test that check_secrets=False allows profiles with secrets."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "secret"),  # Not using env var
            ),
        )

        # Should not raise an error when check_secrets=False
        profile.save(tmp_path, check_secrets=False)
        with pytest.raises(ValueError, match="Profile contains exposed secret keys"):
            profile.save(tmp_path, check_secrets=True)

    def test_multiple_exposed_secrets(self):
        """Test error message when multiple secrets are exposed."""
        profile = Profile(
            con_name="snowflake",
            kwargs_tuple=(
                ("host", "localhost"),
                ("database", "snowflake"),
                ("password", "secret"),  # Not using env var
                ("user", "admin"),  # Not using env var
                ("account", "acc123"),  # Not using env var
            ),
        )

        with pytest.raises(ValueError) as excinfo:
            profile.check_for_exposed_secrets()

        # Check error message contains all secrets
        error_msg = str(excinfo.value)
        assert "'password'" in error_msg
        assert "'user'" in error_msg
        assert "'account'" in error_msg

    def test_unknown_backend_defaults_to_password(self):
        """Test that unknown backends default to checking password."""
        profile = Profile(
            con_name="duckdb",  # Not in the secret_keys dict
            kwargs_tuple=(
                ("path", "mydb.duckdb"),
                ("password", "secret"),  # Not using env var
            ),
        )

        with pytest.raises(ValueError) as excinfo:
            profile.check_for_exposed_secrets()

        # Check error message contains password
        assert "'password'" in str(excinfo.value)

    def test_save_method_calls_check_secrets_fail_then_pass(
        self, tmp_path, monkeypatch
    ):
        """Test that save() method calls _check_for_exposed_secrets."""
        profile = Profile(
            con_name="postgres",
            kwargs_tuple=(
                ("host", "localhost"),
                ("port", 5432),
                ("database", "postgres"),
                ("user", "postgres"),
                ("password", "secret"),  # Not using env var
            ),
        )

        # Override the profile directory for testing
        monkeypatch.setattr("xorq.api.options.profiles.profile_dir", tmp_path)

        with pytest.raises(ValueError) as excinfo:
            profile.save()

        assert "'password'" in str(excinfo.value)

        # Should succeed with check_secrets=False
        profile.save(check_secrets=False)


def test_profile_from_con_preserves_env_vars(monkeypatch, tmp_path):
    """Test that Profile.from_con() preserves environment variables from the original profile."""

    # Set up the profile directory for testing
    monkeypatch.setattr(xo.options.profiles, "profile_dir", tmp_path)

    # Set up environment variables
    monkeypatch.setenv("POSTGRES_HOST", "localhost")
    monkeypatch.setenv("POSTGRES_USER", "postgres")
    monkeypatch.setenv("POSTGRES_PASSWORD", "postgres")

    # Create a profile with environment variable references
    original_profile = Profile(
        con_name="postgres",
        kwargs_tuple=(
            ("host", "${POSTGRES_HOST}"),
            ("port", 5432),
            ("database", "postgres"),
            ("user", "${POSTGRES_USER}"),
            ("password", "${POSTGRES_PASSWORD}"),
        ),
    )

    # Create a connection from the profile
    try:
        connection = original_profile.get_con()

        # Create a profile from the connection using from_con
        profile_from_connection = Profile.from_con(connection)

        # Check if environment variables are preserved
        assert profile_from_connection.kwargs_dict["host"] == "${POSTGRES_HOST}"
        assert profile_from_connection.kwargs_dict["user"] == "${POSTGRES_USER}"
        assert profile_from_connection.kwargs_dict["password"] == "${POSTGRES_PASSWORD}"

        # Test saving and loading the profile from connection
        saved_path = profile_from_connection.save(alias="test_profile", clobber=True)
        assert saved_path.exists()
        loaded_profile = Profile.load("test_profile")

        # Check loaded profile still has env vars
        assert loaded_profile.kwargs_dict["host"] == "${POSTGRES_HOST}"
        assert loaded_profile.kwargs_dict["user"] == "${POSTGRES_USER}"
        assert loaded_profile.kwargs_dict["password"] == "${POSTGRES_PASSWORD}"

    except Exception as e:
        if "connection refused" in str(e).lower():
            import pytest

            pytest.skip(f"Database connection failed: {e}")
        else:
            raise


def test_profile_matches_find_backend(data_dir):
    path = data_dir / "parquet" / "diamonds.parquet"
    con = xo.connect()
    t = xo.deferred_read_parquet(path, con)
    assert con._profile == t._find_backend()._profile



from __future__ import annotations

import math
import sqlite3

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import NULL, AggGen, SQLGlotCompiler
from xorq.vendor.ibis.backends.sql.datatypes import SQLiteType
from xorq.vendor.ibis.backends.sql.dialects import SQLite
from xorq.vendor.ibis.common.temporal import DateUnit, IntervalUnit


class SQLiteCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = SQLite
    type_mapper = SQLiteType
    supports_time_shift_modifiers = sqlite3.sqlite_version_info >= (3, 46, 0)
    supports_subsec = sqlite3.sqlite_version_info >= (3, 42, 0)

    # We could set `supports_order_by=True` for SQLite >= 3.44.0 (2023-11-01).
    agg = AggGen(supports_filter=True)

    NAN = NULL
    POS_INF = sge.Literal.number("1e999")
    NEG_INF = sge.Literal.number("-1e999")

    UNSUPPORTED_OPS = (
        ops.Levenshtein,
        ops.RegexSplit,
        ops.StringSplit,
        ops.IsNan,
        ops.IsInf,
        ops.Covariance,
        ops.Correlation,
        ops.Median,
        ops.ApproxMedian,
        ops.Array,
        ops.ArrayConcat,
        ops.ArrayStringJoin,
        ops.ArrayContains,
        ops.ArrayFlatten,
        ops.ArrayLength,
        ops.ArraySort,
        ops.ArrayStringJoin,
        ops.CountDistinctStar,
        ops.IntervalBinary,
        ops.IntervalAdd,
        ops.IntervalSubtract,
        ops.IntervalMultiply,
        ops.IntervalFloorDivide,
        ops.TimestampBucket,
        ops.TimestampDiff,
        ops.StringToDate,
        ops.StringToTimestamp,
        ops.TimeDelta,
        ops.TimestampDelta,
        ops.TryCast,
    )

    SIMPLE_OPS = {
        ops.Arbitrary: "_ibis_first",
        ops.RegexReplace: "_ibis_regex_replace",
        ops.RegexExtract: "_ibis_regex_extract",
        ops.RegexSearch: "_ibis_regex_search",
        ops.Translate: "_ibis_translate",
        ops.Capitalize: "_ibis_capitalize",
        ops.Reverse: "_ibis_reverse",
        ops.RPad: "_ibis_rpad",
        ops.LPad: "_ibis_lpad",
        ops.Repeat: "_ibis_repeat",
        ops.StringAscii: "_ibis_string_ascii",
        ops.ExtractAuthority: "_ibis_extract_authority",
        ops.ExtractFragment: "_ibis_extract_fragment",
        ops.ExtractHost: "_ibis_extract_host",
        ops.ExtractPath: "_ibis_extract_path",
        ops.ExtractProtocol: "_ibis_extract_protocol",
        ops.ExtractUserInfo: "_ibis_extract_user_info",
        ops.BitwiseXor: "_ibis_xor",
        ops.BitwiseNot: "_ibis_inv",
        ops.TypeOf: "typeof",
        ops.BitOr: "_ibis_bit_or",
        ops.BitAnd: "_ibis_bit_and",
        ops.BitXor: "_ibis_bit_xor",
        ops.Mode: "_ibis_mode",
        ops.Time: "time",
        ops.Date: "date",
    }

    def visit_Log10(self, op, *, arg):
        return self.f.anon.log10(arg)

    def visit_Log2(self, op, *, arg):
        return self.f.anon.log2(arg)

    def visit_Log(self, op, *, arg, base):
        func = self.f.anon.log
        if base is None:
            base = math.e
        return func(base, arg)

    def visit_Cast(self, op, *, arg, to) -> sge.Cast:
        if to.is_timestamp():
            if to.timezone not in (None, "UTC"):
                raise com.UnsupportedOperationError(
                    "SQLite does not support casting to timezones other than 'UTC'"
                )
            if op.arg.dtype.is_numeric():
                return self.f.datetime(arg, "unixepoch")
            else:
                return self.f.strftime("%Y-%m-%d %H:%M:%f", arg)
        elif to.is_date():
            return self.f.date(arg)
        elif to.is_time():
            return self.f.time(arg)
        return super().visit_Cast(op, arg=arg, to=to)

    def visit_Limit(self, op, *, parent, n, offset):
        # SQLite doesn't support compiling an OFFSET without a LIMIT, but
        # treats LIMIT -1 as no limit
        return super().visit_Limit(
            op, parent=parent, n=(-1 if n is None else n), offset=offset
        )

    def visit_WindowBoundary(self, op, *, value, preceding):
        if op.value.dtype.is_interval():
            raise com.OperationNotDefinedError(
                "Interval window bounds not supported by SQLite"
            )
        return super().visit_WindowBoundary(op, value=value, preceding=preceding)

    def visit_JoinLink(self, op, **kwargs):
        if op.how == "asof":
            raise com.UnsupportedOperationError(
                "ASOF joins are not supported by SQLite"
            )
        return super().visit_JoinLink(op, **kwargs)

    def visit_StartsWith(self, op, *, arg, start):
        return arg.like(self.f.concat(start, "%"))

    def visit_EndsWith(self, op, *, arg, end):
        return arg.like(self.f.concat("%", end))

    def visit_StrRight(self, op, *, arg, nchars):
        return self.f.substr(arg, -nchars, nchars)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if op.end is not None:
            raise NotImplementedError("`end` not yet implemented")

        if op.start is not None:
            arg = self.f.substr(arg, start + 1)
            pos = self.f.instr(arg, substr)
            return sg.case().when(pos > 0, pos + start).else_(0)

        return self.f.instr(arg, substr)

    def visit_StringJoin(self, op, *, arg, sep):
        args = [arg[0]]
        for item in arg[1:]:
            args.extend([sep, item])
        return self.f.concat(*args)

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.instr(haystack, needle) >= 1

    def visit_ExtractQuery(self, op, *, arg, key):
        if op.key is None:
            return self.f._ibis_extract_full_query(arg)
        return self.f._ibis_extract_query(arg, key)

    def visit_Greatest(self, op, *, arg):
        return self.f.max(*arg)

    def visit_Least(self, op, *, arg):
        return self.f.min(*arg)

    def visit_IdenticalTo(self, op, *, left, right):
        return sge.Is(this=left, expression=right)

    def visit_Clip(self, op, *, arg, lower, upper):
        if upper is not None:
            arg = self.if_(arg.is_(NULL), arg, self.f.min(upper, arg))

        if lower is not None:
            arg = self.if_(arg.is_(NULL), arg, self.f.max(lower, arg))

        return arg

    def visit_RandomScalar(self, op, **kwargs):
        return 0.5 + self.f.random() / sge.Literal.number(float(-1 << 64))

    def visit_Cot(self, op, *, arg):
        return 1 / self.f.tan(arg)

    def visit_ArgMin(self, *args, **kwargs):
        return self._visit_arg_reduction("min", *args, **kwargs)

    def visit_ArgMax(self, *args, **kwargs):
        return self._visit_arg_reduction("max", *args, **kwargs)

    def _visit_arg_reduction(self, func, op, *, arg, key, where):
        cond = arg.is_(sg.not_(NULL))

        if op.where is not None:
            cond = sg.and_(cond, where)

        agg = self.agg[func](key, where=cond)
        return self.f.anon.json_extract(self.f.json_array(arg, agg), "$[0]")

    def visit_UnwrapJSONString(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("text")),
            self.f.json_extract_scalar(arg, "$"),
            NULL,
        )

    def visit_UnwrapJSONInt64(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("integer")),
            self.cast(self.f.json_extract_scalar(arg, "$"), op.dtype),
            NULL,
        )

    def visit_UnwrapJSONFloat64(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).isin(sge.convert("integer"), sge.convert("real")),
            self.cast(self.f.json_extract_scalar(arg, "$"), op.dtype),
            NULL,
        )

    def visit_UnwrapJSONBoolean(self, op, *, arg):
        return self.if_(
            # isin doesn't work here, with a strange error from sqlite about a
            # misused row value
            self.f.json_type(arg).isin(sge.convert("true"), sge.convert("false")),
            self.cast(self.f.json_extract_scalar(arg, "$"), dt.int64),
            NULL,
        )

    def visit_First(self, op, *, arg, where, order_by, include_null):
        func = "_ibis_first_include_null" if include_null else "_ibis_first"
        return self.agg[func](arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        func = "_ibis_last_include_null" if include_null else "_ibis_last"
        return self.agg[func](arg, where=where, order_by=order_by)

    def visit_Variance(self, op, *, arg, how, where):
        return self.agg[f"_ibis_var_{op.how}"](arg, where=where)

    def visit_StandardDev(self, op, *, arg, how, where):
        var = self.agg[f"_ibis_var_{op.how}"](arg, where=where)
        return self.f.sqrt(var)

    def visit_ApproxCountDistinct(self, op, *, arg, where):
        return self.agg.count(sge.Distinct(expressions=[arg]), where=where)

    def visit_CountDistinct(self, op, *, arg, where):
        return self.agg.count(sge.Distinct(expressions=[arg]), where=where)

    def visit_Strftime(self, op, *, arg, format_str):
        return self.f.strftime(format_str, arg)

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.date(self.f.printf("%04d-%02d-%02d", year, month, day))

    def visit_TimeFromHMS(self, op, *, hours, minutes, seconds):
        return self.f.time(self.f.printf("%02d:%02d:%02d", hours, minutes, seconds))

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        return self.f.datetime(
            self.f.printf(
                "%04d-%02d-%02d %02d:%02d:%02d%s",
                year,
                month,
                day,
                hours,
                minutes,
                seconds,
            )
        )

    def visit_Modulus(self, op, *, left, right):
        return self.f.anon.mod(left, right)

    def _temporal_truncate(self, func, arg, unit):
        if unit.short == "Q":
            return sge.Case(
                ifs=[
                    self.if_(
                        sge.Between(
                            this=self.cast(self.f.strftime("%m", arg), dt.int32),
                            low=sge.convert(lower),
                            high=sge.convert(lower + 2),
                        ),
                        self.f.strftime(f"%Y-{lower:0>2}-01", arg),
                    )
                    for lower in range(1, 13, 3)
                ],
            )
        modifiers = {
            DateUnit.DAY: ("start of day",),
            DateUnit.WEEK: ("weekday 0", "-6 days"),
            DateUnit.MONTH: ("start of month",),
            DateUnit.YEAR: ("start of year",),
            IntervalUnit.DAY: ("start of day",),
            IntervalUnit.WEEK: ("weekday 0", "-6 days", "start of day"),
            IntervalUnit.MONTH: ("start of month",),
            IntervalUnit.YEAR: ("start of year",),
        }

        params = modifiers.get(unit)
        if params is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit}")
        return func(arg, *params)

    def visit_DateTruncate(self, op, *, arg, unit):
        return self._temporal_truncate(self.f.date, arg, unit)

    def visit_TimestampTruncate(self, op, *, arg, unit):
        return self._temporal_truncate(self.f.anon.datetime, arg, unit)

    def visit_DateArithmetic(self, op, *, left, right):
        right = right.this

        if (unit := op.right.dtype.unit) in (
            IntervalUnit.QUARTER,
            IntervalUnit.MICROSECOND,
            IntervalUnit.NANOSECOND,
        ):
            raise com.UnsupportedOperationError(
                f"SQLite does not support `{unit}` units in temporal arithmetic"
            )
        elif unit == IntervalUnit.WEEK:
            unit = IntervalUnit.DAY
            right *= 7
        elif unit == IntervalUnit.MILLISECOND:
            # sqlite doesn't allow milliseconds, so divide milliseconds by 1e3 to
            # get seconds, and change the unit to seconds
            unit = IntervalUnit.SECOND
            right /= 1e3

        # compute whether we're adding or subtracting an interval
        sign = "+" if isinstance(op, (ops.DateAdd, ops.TimestampAdd)) else "-"

        modifiers = []

        # floor the result if the unit is a year, month, or day to match other
        # backend behavior
        if unit in (IntervalUnit.YEAR, IntervalUnit.MONTH, IntervalUnit.DAY):
            if not self.supports_time_shift_modifiers:
                raise com.UnsupportedOperationError(
                    "SQLite does not support time shift modifiers until version 3.46; "
                    f"found version {sqlite3.sqlite_version}"
                )
            modifiers.append("floor")

        if isinstance(op, (ops.TimestampAdd, ops.TimestampSub)):
            # if the left operand is a timestamp, return as much precision as
            # possible
            if not self.supports_subsec:
                raise com.UnsupportedOperationError(
                    "SQLite does not support subsecond resolution until version 3.42; "
                    f"found version {sqlite3.sqlite_version}"
                )
            func = self.f.datetime
            modifiers.append("subsec")
        else:
            func = self.f.date

        return func(
            left,
            self.f.concat(
                sign, self.cast(right, dt.string), " ", unit.singular.lower()
            ),
            *modifiers,
            dialect=self.dialect,
        )

    visit_TimestampAdd = visit_TimestampSub = visit_DateAdd = visit_DateSub = (
        visit_DateArithmetic
    )

    def visit_DateDiff(self, op, *, left, right):
        return self.f.julianday(left) - self.f.julianday(right)

    def visit_ExtractYear(self, op, *, arg):
        return self.cast(self.f.strftime("%Y", arg), dt.int64)

    def visit_ExtractQuarter(self, op, *, arg):
        return (self.f.strftime("%m", arg) + 2) / 3

    def visit_ExtractMonth(self, op, *, arg):
        return self.cast(self.f.strftime("%m", arg), dt.int64)

    def visit_ExtractDay(self, op, *, arg):
        return self.cast(self.f.strftime("%d", arg), dt.int64)

    def visit_ExtractDayOfYear(self, op, *, arg):
        return self.cast(self.f.strftime("%j", arg), dt.int64)

    def visit_ExtractHour(self, op, *, arg):
        return self.cast(self.f.strftime("%H", arg), dt.int64)

    def visit_ExtractMinute(self, op, *, arg):
        return self.cast(self.f.strftime("%M", arg), dt.int64)

    def visit_ExtractSecond(self, op, *, arg):
        return self.cast(self.f.strftime("%S", arg), dt.int64)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.cast(self.f.mod(self.f.strftime("%f", arg) * 1000, 1000), dt.int64)

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.cast(
            self.f.mod(self.cast(self.f.strftime("%f", arg), dt.int64), 1000), dt.int64
        )

    def visit_ExtractWeekOfYear(self, op, *, arg):
        """ISO week of year.

        This solution is based on https://stackoverflow.com/a/15511864 and handle
        the edge cases when computing ISO week from non-ISO week.

        The implementation gives the same results as `datetime.isocalendar()`.

        The year's week that "wins" the day is the year with more allotted days.

        For example:

        ```
        $ cal '2011-01-01'
            January 2011
        Su Mo Tu We Th Fr Sa
                        |1|
        2  3  4  5  6  7  8
        9 10 11 12 13 14 15
        16 17 18 19 20 21 22
        23 24 25 26 27 28 29
        30 31
        ```

        Here the ISO week number is `52` since the day occurs in a week with more
        days in the week occurring in the _previous_ week's year.

        ```
        $ cal '2012-12-31'
            December 2012
        Su Mo Tu We Th Fr Sa
                        1
        2  3  4  5  6  7  8
        9 10 11 12 13 14 15
        16 17 18 19 20 21 22
        23 24 25 26 27 28 29
        30 |31|
        ```

        Here the ISO week of year is `1` since the day occurs in a week with more
        days in the week occurring in the _next_ week's year.
        """
        date = self.f.date(arg, "-3 days", "weekday 4")
        return (self.f.strftime("%j", date) - 1) / 7 + 1

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.cast(self.f.strftime("%s", arg), dt.int64)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return self.cast(
            self.f.mod(self.cast(self.f.strftime("%w", arg) + 6, dt.int64), 7), dt.int64
        )

    def visit_DayOfWeekName(self, op, *, arg):
        return sge.Case(
            this=self.f.strftime("%w", arg),
            ifs=[
                self.if_("0", "Sunday"),
                self.if_("1", "Monday"),
                self.if_("2", "Tuesday"),
                self.if_("3", "Wednesday"),
                self.if_("4", "Thursday"),
                self.if_("5", "Friday"),
                self.if_("6", "Saturday"),
            ],
        )

    def visit_Xor(self, op, *, left, right):
        return (left.or_(right)).and_(sg.not_(left.and_(right)))

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.f.unhex(value.hex())

        if dtype.is_decimal():
            value = float(value)
            dtype = dt.double(nullable=dtype.nullable)
        elif dtype.is_uuid():
            value = str(value)
            dtype = dt.string(nullable=dtype.nullable)
        elif dtype.is_interval():
            value = int(value)
            dtype = dt.int64(nullable=dtype.nullable)
        elif dtype.is_date() or dtype.is_timestamp() or dtype.is_time():
            # To ensure comparisons apply uniformly between temporal values
            # (which are always represented as strings), we need to enforce
            # that temporal literals are formatted the same way that SQLite
            # formats them. This means " " instead of "T" and no offset suffix
            # for UTC.
            value = (
                value.isoformat()
                .replace("T", " ")
                .replace("Z", "")
                .replace("+00:00", "")
            )
            dtype = dt.string(nullable=dtype.nullable)
        elif (
            dtype.is_map()
            or dtype.is_struct()
            or dtype.is_array()
            or dtype.is_geospatial()
        ):
            raise com.UnsupportedBackendType(f"Unsupported type: {dtype!r}")
        return super().visit_NonNullLiteral(op, value=value, dtype=dtype)

    def visit_DateDelta(self, op, *, left, right, part):
        if not isinstance(part, sge.Literal):
            raise com.UnsupportedOperationError(
                "Only literal `part` values are supported for date delta"
            )
        if part.this != "day":
            raise com.UnsupportedOperationError(
                f"Only 'day' part is supported for date delta in the {self.dialect} backend"
            )
        return self.f._ibis_date_delta(left, right)

    def visit_Hash(self, op, *, arg):
        return self.f.anon.city_hash_32(arg)


compiler = SQLiteCompiler()



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from xorq.vendor.ibis.util import PseudoHashable, indent


if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    import pyarrow as pa

    from xorq.vendor.ibis.expr.datatypes import DataType
    from xorq.vendor.ibis.expr.schema import Schema

C = TypeVar("C")
T = TypeVar("T")
S = TypeVar("S")


class TypeMapper(Generic[T]):
    # `T` is the format-specific type object, e.g. pyarrow.DataType

    @classmethod
    def from_ibis(cls, dtype: DataType) -> T:
        """Convert an Ibis DataType to a format-specific type object.

        Parameters
        ----------
        dtype
            The Ibis DataType to convert.

        Returns
        -------
        Format-specific type object.

        """
        raise NotImplementedError

    @classmethod
    def to_ibis(cls, typ: T, nullable: bool = True) -> DataType:
        """Convert a format-specific type object to an Ibis DataType.

        Parameters
        ----------
        typ
            The format-specific type object to convert.
        nullable
            Whether the Ibis DataType should be nullable.

        Returns
        -------
        Ibis DataType.

        """
        raise NotImplementedError

    @classmethod
    def from_string(cls, text: str, nullable: bool = True) -> DataType:
        """Convert a backend-specific string representation into an Ibis DataType.

        Parameters
        ----------
        text
            The backend-specific string representation to convert.
        nullable
            Whether the Ibis DataType should be nullable.

        Returns
        -------
        Ibis DataType.

        """
        raise NotImplementedError

    @classmethod
    def to_string(cls, dtype: DataType) -> str:
        """Convert `dtype` into a backend-specific string representation.

        Parameters
        ----------
        dtype
            The Ibis DataType to convert.

        Returns
        -------
        Backend-specific string representation.

        """
        raise NotImplementedError


class SchemaMapper(Generic[S]):
    # `S` is the format-specific schema object, e.g. pyarrow.Schema

    @classmethod
    def from_ibis(cls, schema: Schema) -> S:
        """Convert an Ibis Schema to a format-specific schema object.

        Parameters
        ----------
        schema
            The Ibis Schema to convert.

        Returns
        -------
        Format-specific schema object.

        """
        raise NotImplementedError

    @classmethod
    def to_ibis(cls, obj: S) -> Schema:
        """Convert a format-specific schema object to an Ibis Schema.

        Parameters
        ----------
        obj
            The format-specific schema object to convert.

        Returns
        -------
        Ibis Schema.

        """
        raise NotImplementedError


class DataMapper(Generic[S, C, T]):
    # `S` is the format-specific scalar object, e.g. pyarrow.Scalar
    # `C` is the format-specific column object, e.g. pyarrow.Array
    # `T` is the format-specific table object, e.g. pyarrow.Table

    @classmethod
    def convert_scalar(cls, obj: S, dtype: DataType) -> S:
        """Convert a format-specific scalar to the given ibis datatype.

        Parameters
        ----------
        obj
            The format-specific scalar value to convert.
        dtype
            The Ibis datatype to convert to.

        Returns
        -------
        Format specific scalar corresponding to the given Ibis datatype.

        """
        raise NotImplementedError

    @classmethod
    def convert_column(cls, obj: C, dtype: DataType) -> C:
        """Convert a format-specific column to the given ibis datatype.

        Parameters
        ----------
        obj
            The format-specific column value to convert.
        dtype
            The Ibis datatype to convert to.

        Returns
        -------
        Format specific column corresponding to the given Ibis datatype.

        """
        raise NotImplementedError

    @classmethod
    def convert_table(cls, obj: T, schema: Schema) -> T:
        """Convert a format-specific table to the given ibis schema.

        Parameters
        ----------
        obj
            The format-specific table-like object to convert.
        schema
            The Ibis schema to convert to.

        Returns
        -------
        Format specific table-like object corresponding to the given Ibis schema.

        """
        raise NotImplementedError

    @classmethod
    def infer_scalar(cls, obj: S) -> DataType:
        """Infer the Ibis datatype of a format-specific scalar.

        Parameters
        ----------
        obj
            The format-specific scalar to infer the Ibis datatype of.

        Returns
        -------
        Ibis datatype corresponding to the given format-specific scalar.

        """
        raise NotImplementedError

    @classmethod
    def infer_column(cls, obj: C) -> DataType:
        """Infer the Ibis datatype of a format-specific column.

        Parameters
        ----------
        obj
            The format-specific column to infer the Ibis datatype of.

        Returns
        -------
        Ibis datatype corresponding to the given format-specific column.

        """
        raise NotImplementedError

    @classmethod
    def infer_table(cls, obj: T) -> Schema:
        """Infer the Ibis schema of a format-specific table.

        Parameters
        ----------
        obj
            The format-specific table to infer the Ibis schema of.

        Returns
        -------
        Ibis schema corresponding to the given format-specific table.

        """
        raise NotImplementedError


class TableProxy(PseudoHashable[T]):
    def __repr__(self) -> str:
        data_repr = indent(repr(self.obj), spaces=2)
        return f"{self.__class__.__name__}:\n{data_repr}"

    def __len__(self) -> int:
        return len(self.obj)

    @abstractmethod
    def proxy_equals(self, other: "TableProxy") -> bool:  # pragma: no cover
        """Return True if the underlying data compares equal to other's data."""

    @abstractmethod
    def to_frame(self) -> pd.DataFrame:  # pragma: no cover
        """Convert this input to a pandas DataFrame."""

    @abstractmethod
    def to_pyarrow(self, schema: Schema) -> pa.Table:  # pragma: no cover
        """Convert this input to a PyArrow Table."""

    @abstractmethod
    def to_polars(self, schema: Schema) -> pl.DataFrame:  # pragma: no cover
        """Convert this input to a Polars DataFrame."""

    def to_pyarrow_bytes(self, schema: Schema) -> bytes:
        import pyarrow as pa
        import pyarrow_hotfix  # noqa: F401

        data = self.to_pyarrow(schema=schema)
        out = pa.BufferOutputStream()
        with pa.RecordBatchFileWriter(out, data.schema) as writer:
            writer.write(data)
        return out.getvalue()



from __future__ import annotations

import random

import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.types as ir
from xorq.api import _
from xorq.tests.util import assert_frame_equal


@pytest.fixture
def union_subsets(alltypes, alltypes_df):
    cols_a, cols_b, cols_c = (alltypes.columns.copy() for _ in range(3))

    random.seed(89)
    random.shuffle(cols_a)
    random.shuffle(cols_b)
    random.shuffle(cols_c)
    assert cols_a != cols_b != cols_c

    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))[cols_a]
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))[cols_b]
    c = alltypes.filter((_.id >= 5213) & (_.id <= 5220))[cols_c]

    da = alltypes_df[(alltypes_df.id >= 5200) & (alltypes_df.id <= 5210)][cols_a]
    db = alltypes_df[(alltypes_df.id >= 5205) & (alltypes_df.id <= 5215)][cols_b]
    dc = alltypes_df[(alltypes_df.id >= 5213) & (alltypes_df.id <= 5220)][cols_c]

    return (a, b, c), (da, db, dc)


@pytest.mark.parametrize("distinct", [False, True], ids=["all", "distinct"])
def test_union(union_subsets, distinct):
    (a, b, c), (da, db, dc) = union_subsets

    expr = xo.union(a, b, distinct=distinct).order_by("id")
    result = expr.execute()

    expected = pd.concat([da, db], axis=0).sort_values("id").reset_index(drop=True)
    if distinct:
        expected = expected.drop_duplicates("id")

    assert_frame_equal(result, expected)


def test_union_mixed_distinct(union_subsets):
    (a, b, c), (da, db, dc) = union_subsets

    expr = a.union(b, distinct=True).union(c, distinct=False).order_by("id")
    result = expr.execute()
    expected = pd.concat(
        [pd.concat([da, db], axis=0).drop_duplicates("id"), dc], axis=0
    ).sort_values("id")

    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_intersect(alltypes, alltypes_df, distinct):
    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))
    c = alltypes.filter((_.id >= 5195) & (_.id <= 5208))

    # Reset index to ensure simple RangeIndex, needed for computing `expected`
    df = alltypes_df.reset_index(drop=True)
    da = alltypes_df[(alltypes_df.id >= 5200) & (alltypes_df.id <= 5210)]
    db = alltypes_df[(alltypes_df.id >= 5205) & (alltypes_df.id <= 5215)]
    dc = alltypes_df[(alltypes_df.id >= 5195) & (alltypes_df.id <= 5208)]

    expr = xo.intersect(a, b, c, distinct=distinct).order_by("id")
    result = expr.execute()

    index = da.index.intersection(db.index).intersection(dc.index)
    expected = df.iloc[index].sort_values("id").reset_index(drop=True)
    if distinct:
        expected = expected.drop_duplicates()

    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_difference(alltypes, alltypes_df, distinct):
    a = alltypes.filter((_.id >= 5200) & (_.id <= 5210))
    b = alltypes.filter((_.id >= 5205) & (_.id <= 5215))
    c = alltypes.filter((_.id >= 5195) & (_.id <= 5202))

    # Reset index to ensure simple RangeIndex, needed for computing `expected`
    df = alltypes_df.reset_index(drop=True)
    da = alltypes_df[(alltypes_df.id >= 5200) & (alltypes_df.id <= 5210)]
    db = alltypes_df[(alltypes_df.id >= 5205) & (alltypes_df.id <= 5215)]
    dc = alltypes_df[(alltypes_df.id >= 5195) & (alltypes_df.id <= 5202)]

    expr = xo.difference(a, b, c, distinct=distinct).order_by("id")
    result = expr.execute()

    index = da.index.difference(db.index).difference(dc.index)
    expected = df.iloc[index].sort_values("id").reset_index(drop=True)
    if distinct:
        expected = expected.drop_duplicates()

    assert_frame_equal(result, expected)


@pytest.mark.parametrize("method", ["intersect", "difference", "union"])
def test_table_set_operations_api(alltypes, method):
    # top level variadic
    result = getattr(xo, method)(alltypes)
    assert result.equals(alltypes)

    # table level methods require at least one argument
    with pytest.raises(
        TypeError, match="missing 1 required positional argument: 'table'"
    ):
        getattr(ir.Table, method)(alltypes)


@pytest.mark.parametrize(
    "distinct",
    [
        True,
        False,
    ],
)
def test_top_level_union(con, alltypes, distinct):
    t1 = alltypes.select(a="bigint_col").filter(lambda t: t.a == 10).distinct()
    t2 = alltypes.select(a="bigint_col").filter(lambda t: t.a == 20).distinct()
    expr = t1.union(t2, distinct=distinct).limit(2)
    result = con.execute(expr)
    expected = pd.DataFrame({"a": [10, 20]})
    assert_frame_equal(result.sort_values("a").reset_index(drop=True), expected)


@pytest.mark.parametrize(
    "distinct",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    ("opname", "expected"),
    [
        ("intersect", pd.DataFrame({"a": [20]})),
        ("difference", pd.DataFrame({"a": [10]})),
    ],
    ids=["intersect", "difference"],
)
def test_top_level_intersect_difference(con, alltypes, distinct, opname, expected):
    t1 = (
        alltypes.select(a="bigint_col")
        .filter(lambda t: (t.a == 10) | (t.a == 20))
        .distinct()
    )
    t2 = (
        alltypes.select(a="bigint_col")
        .filter(lambda t: (t.a == 20) | (t.a == 30))
        .distinct()
    )
    op = getattr(t1, opname)
    expr = op(t2, distinct=distinct).limit(2)
    result = con.execute(expr)
    assert_frame_equal(result, expected)
